"""The range dialect for Python.

This dialect models the builtin `range()` function in Python.

The dialect includes:
- The `Range` statement class.
- The lowering pass for the `range()` function.

This dialect does not include a concrete implementation or type inference
for the `range()` function. One needs to use other dialect for the concrete
implementation and type inference, e.g., `ilist` dialect.
"""

import ast
from dataclasses import dataclass

from kirin import ir, types, interp, lowering, exceptions
from kirin.decl import info, statement
from kirin.dialects import eltype

dialect = ir.Dialect("py.range")


@dataclass(frozen=True)
class RangeLowering(ir.FromPythonCall["Range"]):

    def lower(
        self, stmt: type["Range"], state: lowering.LoweringState, node: ast.Call
    ) -> lowering.Result:
        return _lower_range(state, node)


@statement(dialect=dialect)
class Range(ir.Statement):
    name = "range"
    traits = frozenset({ir.Pure(), RangeLowering()})
    start: ir.SSAValue = info.argument(types.Int)
    stop: ir.SSAValue = info.argument(types.Int)
    step: ir.SSAValue = info.argument(types.Int)
    result: ir.ResultValue = info.result(types.PyClass(range))


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Call_range(
        self, state: lowering.LoweringState, node: ast.Call
    ) -> lowering.Result:
        return _lower_range(state, node)


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(eltype.ElType, types.PyClass(range))
    def eltype_range(self, interp_, frame: interp.Frame, stmt: eltype.ElType):
        return (types.Int,)


def _lower_range(state: lowering.LoweringState, node: ast.Call) -> lowering.Result:
    if len(node.args) == 1:
        start = state.visit(ast.Constant(0)).expect_one()
        stop = state.visit(node.args[0]).expect_one()
        step = state.visit(ast.Constant(1)).expect_one()
    elif len(node.args) == 2:
        start = state.visit(node.args[0]).expect_one()
        stop = state.visit(node.args[1]).expect_one()
        step = state.visit(ast.Constant(1)).expect_one()
    elif len(node.args) == 3:
        start = state.visit(node.args[0]).expect_one()
        stop = state.visit(node.args[1]).expect_one()
        step = state.visit(node.args[2]).expect_one()
    else:
        raise exceptions.DialectLoweringError("range() takes 1-3 arguments")

    return lowering.Result(state.append_stmt(Range(start, stop, step)))



from typing import Literal

from kirin import ir, types, rewrite
from kirin.passes import aggressive
from kirin.prelude import python_basic
from kirin.dialects import py, func, ilist, lowering
from kirin.passes.typeinfer import TypeInfer


@ir.dialect_group(python_basic.union([func, ilist, lowering.func]))
def basic(self):
    aggressive_fold_pass = aggressive.Fold(self)
    typeinfer_pass = TypeInfer(self)

    def run_pass(
        mt: ir.Method,
    ) -> None:
        aggressive_fold_pass.fixpoint(mt)
        rewrite.Fixpoint(rewrite.Walk(ilist.rewrite.ConstList2IList())).rewrite(mt.code)
        typeinfer_pass(mt)

    return run_pass


def test_empty():
    @basic
    def empty_list():
        return []

    empty_list.print()
    assert empty_list.return_type.is_subseteq(ilist.IListType[types.Any])


def test_typehint():
    @basic
    def main(xs: ilist.IList[int, Literal[3]]):
        return xs + [4, 5, 6] + xs

    assert main.return_type is not None
    assert main.return_type.is_subseteq(ilist.IListType[types.Int, types.Literal(9)])


@basic
def add1(x: int):
    return x + 1


def test_ilist_fcf():
    # TODO: actually check equivalent code
    rule = rewrite.Fixpoint(rewrite.Walk(ilist.rewrite.Unroll()))

    xs = ilist.IList([1, 2, 3])

    @basic
    def map(xs: ilist.IList[int, Literal[3]]):
        return ilist.map(add1, xs)

    @basic
    def foreach(xs: ilist.IList[int, Literal[3]]):
        ilist.for_each(add1, xs)

    map_before = map(xs)
    foreach_before = foreach(xs)
    rule.rewrite(map.code)
    rule.rewrite(foreach.code)
    map_after = map(xs)
    foreach_after = foreach(xs)
    assert map_before.data == map_after.data  # type: ignore
    assert foreach_before == foreach_after

    assert isinstance(map.callable_region.blocks[0].stmts.at(1), py.Constant)
    assert isinstance(map.callable_region.blocks[0].stmts.at(-2), ilist.New)

    assert isinstance(foreach.callable_region.blocks[0].stmts.at(1), py.Constant)
    assert isinstance(
        foreach.callable_region.blocks[0].stmts.at(2), py.indexing.GetItem
    )
    assert isinstance(foreach.callable_region.blocks[0].stmts.at(3), func.Call)
    assert isinstance(foreach.callable_region.blocks[0].stmts.at(10), func.ConstantNone)

    @basic
    def add(x: int, y: int):
        return x + y, y

    @basic
    def scan(xs: ilist.IList[int, Literal[3]]):
        return ilist.Scan(add, xs, init=123)  # type: ignore

    scan_before = scan(xs)
    rule.rewrite(scan.code)
    scan_after = scan(xs)
    assert scan_before == scan_after  # type: ignore
    assert isinstance(scan.callable_region.blocks[0].stmts.at(-2), py.tuple.New)
    assert isinstance(scan.callable_region.blocks[0].stmts.at(-3), ilist.New)

    @basic
    def add2(x: int, y: int):
        return x + y

    @basic
    def foldl(xs: ilist.IList[int, Literal[3]]):
        return ilist.Foldl(add2, xs, init=123)  # type: ignore

    @basic
    def foldr(xs: ilist.IList[int, Literal[3]]):
        return ilist.Foldr(add2, xs, init=123)  # type: ignore

    foldl_before = foldl(xs)
    foldr_before = foldr(xs)
    rule.rewrite(foldl.code)
    rule.rewrite(foldr.code)
    foldl_after = foldl(xs)
    foldr_after = foldr(xs)

    assert foldl_before == foldl_after  # type: ignore
    assert foldr_before == foldr_after  # type: ignore

    stmt = foldl.callable_region.blocks[0].stmts.at(2)
    assert isinstance(stmt, py.Constant)
    assert stmt.value == 0
    assert isinstance(foldl.callable_region.blocks[0].stmts.at(4), func.Call)

    stmt = foldl.callable_region.blocks[0].stmts.at(5)
    assert isinstance(stmt, py.Constant)
    assert stmt.value == 1
    assert isinstance(foldl.callable_region.blocks[0].stmts.at(7), func.Call)

    stmt = foldl.callable_region.blocks[0].stmts.at(8)
    assert isinstance(stmt, py.Constant)
    assert stmt.value == 2
    assert isinstance(foldl.callable_region.blocks[0].stmts.at(10), func.Call)

    # ========== foldl
    stmt = foldr.callable_region.blocks[0].stmts.at(2)
    assert isinstance(stmt, py.Constant)
    assert stmt.value == 2
    assert isinstance(foldr.callable_region.blocks[0].stmts.at(4), func.Call)

    stmt = foldr.callable_region.blocks[0].stmts.at(5)
    assert isinstance(stmt, py.Constant)
    assert stmt.value == 1
    assert isinstance(foldr.callable_region.blocks[0].stmts.at(7), func.Call)

    stmt = foldr.callable_region.blocks[0].stmts.at(8)
    assert isinstance(stmt, py.Constant)
    assert stmt.value == 0
    assert isinstance(foldr.callable_region.blocks[0].stmts.at(10), func.Call)


def test_ilist_range():
    @basic.add(py.range)
    def map():
        return ilist.Map(add1, range(0, 3))  # type: ignore

    assert map() == ilist.IList([1, 2, 3])

    @basic.add(py.range)
    def const_range():
        return range(0, 3)

    assert const_range() == ilist.IList(range(0, 3))


@basic
def main(xs: ilist.IList[int, Literal[3]]):
    return xs + [4, 5, 6] + xs


main.print()



"""Kirin IR Traits.

This module defines the traits that can be used to define the behavior of
Kirin IR nodes. The base trait is `StmtTrait`, which is a `dataclass` that
implements the `__hash__` and `__eq__` methods.

There are also some basic traits that are provided for convenience, such as
`Pure`, `HasParent`, `ConstantLike`, `IsTerminator`, `NoTerminator`, and
`IsolatedFromAbove`.
"""

from .abc import (
    StmtTrait as StmtTrait,
    RegionTrait as RegionTrait,
    PythonLoweringTrait as PythonLoweringTrait,
)
from .basic import (
    Pure as Pure,
    HasParent as HasParent,
    MaybePure as MaybePure,
    ConstantLike as ConstantLike,
    IsTerminator as IsTerminator,
    NoTerminator as NoTerminator,
    IsolatedFromAbove as IsolatedFromAbove,
)
from .symbol import SymbolTable as SymbolTable, SymbolOpInterface as SymbolOpInterface
from .callable import (
    HasSignature as HasSignature,
    CallableStmtInterface as CallableStmtInterface,
)
from .lowering.call import FromPythonCall as FromPythonCall
from .region.ssacfg import SSACFGRegion as SSACFGRegion
from .lowering.context import (
    FromPythonWith as FromPythonWith,
    FromPythonWithSingleItem as FromPythonWithSingleItem,
)



"""Lowering Python AST to cf dialect.
"""

import ast

from kirin import ir, types
from kirin.dialects import cf, py
from kirin.lowering import Frame, Result, FromPythonAST, LoweringState
from kirin.exceptions import DialectLoweringError

dialect = ir.Dialect("lowering.cf")


@dialect.register
class CfLowering(FromPythonAST):

    def lower_Pass(self, state: LoweringState, node: ast.Pass) -> Result:
        state.append_stmt(
            cf.Branch(arguments=(), successor=state.current_frame.next_block)
        )
        return Result()

    def lower_For(self, state: LoweringState, node: ast.For) -> Result:
        yields: list[str] = []

        def new_block_arg_if_inside_loop(frame: Frame, capture: ir.SSAValue):
            if not capture.name:
                raise DialectLoweringError("unexpected loop variable captured")
            yields.append(capture.name)
            return frame.entr_block.args.append_from(capture.type, capture.name)

        frame = state.current_frame
        iterable = state.visit(node.iter).expect_one()
        iter_stmt = frame.append_stmt(py.iterable.Iter(iterable))
        none_stmt = frame.append_stmt(py.Constant(None))

        body_frame = state.push_frame(
            Frame.from_stmts(
                node.body,
                state,
                region=state.current_frame.curr_region,
                globals=state.current_frame.globals,
                capture_callback=new_block_arg_if_inside_loop,
            )
        )
        next_value = body_frame.entr_block.args.append_from(types.Any, "next_value")
        py.unpack.unpacking(state, node.target, next_value)
        state.exhaust(body_frame)
        self.branch_next_if_not_terminated(body_frame)
        yield_args = tuple(body_frame.get_scope(name) for name in yields)
        next_stmt = py.iterable.Next(iter_stmt.iter)
        cond_stmt = py.cmp.Is(next_stmt.value, none_stmt.result)
        body_frame.next_block.stmts.append(next_stmt)
        body_frame.next_block.stmts.append(cond_stmt)
        body_frame.next_block.stmts.append(
            cf.ConditionalBranch(
                cond_stmt.result,
                yield_args,
                (next_stmt.value,) + yield_args,
                then_successor=frame.next_block,
                else_successor=body_frame.entr_block,
            )
        )
        state.pop_frame()

        # insert the branch to the entrance of the loop (the code block before loop)
        next_stmt = frame.append_stmt(py.iterable.Next(iter_stmt.iter))
        cond_stmt = frame.append_stmt(py.cmp.Is(next_stmt.value, none_stmt.result))
        yield_args = tuple(frame.get_scope(name) for name in yields)
        frame.append_stmt(
            cf.ConditionalBranch(
                cond_stmt.result,
                yield_args,
                (next_stmt.value,) + yield_args,
                then_successor=frame.next_block,  # empty iterator
                else_successor=body_frame.entr_block,
            )
        )

        frame.jump_next()
        for name, arg in zip(yields, yield_args):
            input = frame.curr_block.args.append_from(arg.type, name)
            frame.defs[name] = input
        return Result()

    def lower_If(self, state: LoweringState, node: ast.If) -> Result:
        cond = state.visit(node.test).expect_one()
        frame = state.current_frame
        before_block = frame.curr_block
        if_frame = state.push_frame(
            Frame.from_stmts(
                node.body,
                state,
                region=frame.curr_region,
                globals=frame.globals,
            )
        )
        true_cond = if_frame.entr_block.args.append_from(types.Bool, cond.name)
        if cond.name:
            if_frame.defs[cond.name] = true_cond
        state.exhaust()
        self.branch_next_if_not_terminated(if_frame)
        state.pop_frame()

        else_frame = state.push_frame(
            Frame.from_stmts(
                node.orelse,
                state,
                region=frame.curr_region,
                globals=frame.globals,
            )
        )
        true_cond = else_frame.entr_block.args.append_from(types.Bool, cond.name)
        if cond.name:
            else_frame.defs[cond.name] = true_cond
        state.exhaust()
        self.branch_next_if_not_terminated(else_frame)
        state.pop_frame()

        after_frame = state.push_frame(
            Frame.from_stmts(
                frame.stream.split(),
                state,
                region=frame.curr_region,
                globals=frame.globals,
            )
        )

        after_frame.defs.update(frame.defs)
        phi: set[str] = set()
        for name in if_frame.defs.keys():
            if frame.get(name):
                phi.add(name)
            elif name in else_frame.defs:
                phi.add(name)

        for name in else_frame.defs.keys():
            if frame.get(name):  # not defined in if_frame
                phi.add(name)

        for name in phi:
            after_frame.defs[name] = after_frame.entr_block.args.append_from(
                types.Any, name
            )

        state.exhaust()
        self.branch_next_if_not_terminated(after_frame)
        after_frame.next_block.stmts.append(
            cf.Branch(arguments=(), successor=frame.next_block)
        )
        state.pop_frame()

        if_args = []
        for name in phi:
            if value := if_frame.get(name):
                if_args.append(value)
            else:
                raise DialectLoweringError(f"undefined variable {name} in if branch")

        else_args = []
        for name in phi:
            if value := else_frame.get(name):
                else_args.append(value)
            else:
                raise DialectLoweringError(f"undefined variable {name} in else branch")

        if_frame.next_block.stmts.append(
            cf.Branch(
                arguments=tuple(if_args),
                successor=after_frame.entr_block,
            )
        )
        else_frame.next_block.stmts.append(
            cf.Branch(
                arguments=tuple(else_args),
                successor=after_frame.entr_block,
            )
        )
        before_block.stmts.append(
            cf.ConditionalBranch(
                cond=cond,
                then_arguments=(cond,),
                then_successor=if_frame.entr_block,
                else_arguments=(cond,),
                else_successor=else_frame.entr_block,
            )
        )
        frame.jump_next()
        return Result()

    def branch_next_if_not_terminated(self, frame: Frame):
        """Branch to the next block if the current block is not terminated.

        This must be used after exhausting the current frame and before popping the frame.
        """
        if not frame.curr_block.last_stmt or not frame.curr_block.last_stmt.has_trait(
            ir.IsTerminator
        ):
            frame.curr_block.stmts.append(
                cf.Branch(arguments=(), successor=frame.next_block)
            )

    def current_block_terminated(self, frame: Frame):
        return frame.curr_block.last_stmt and frame.curr_block.last_stmt.has_trait(
            ir.IsTerminator
        )



# TODO: replace with something faster
from typing import Any, Generic, TypeVar, overload
from dataclasses import dataclass
from collections.abc import Sequence

T = TypeVar("T")
L = TypeVar("L")


@dataclass
class IList(Generic[T, L]):
    """A simple immutable list."""

    data: Sequence[T]

    def __hash__(self) -> int:
        return id(self)  # do not hash the data

    def __len__(self) -> int:
        return len(self.data)

    @overload
    def __add__(self, other: "IList[T, Any]") -> "IList[T, Any]": ...

    @overload
    def __add__(self, other: list[T]) -> "IList[T, Any]": ...

    def __add__(self, other):
        return IList(self.data + other)

    @overload
    def __radd__(self, other: "IList[T, Any]") -> "IList[T, Any]": ...

    @overload
    def __radd__(self, other: list[T]) -> "IList[T, Any]": ...

    def __radd__(self, other):
        return IList(other + self.data)

    def __repr__(self) -> str:
        return f"IList({self.data})"

    def __str__(self) -> str:
        return f"IList({self.data})"

    def __iter__(self):
        return iter(self.data)

    @overload
    def __getitem__(self, index: slice) -> "IList[T, Any]": ...

    @overload
    def __getitem__(self, index: int) -> T: ...

    def __getitem__(self, index: int | slice) -> T | "IList[T, Any]":
        if isinstance(index, slice):
            return IList(self.data[index])
        return self.data[index]

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, IList):
            return False
        return self.data == value.data
